/*
===============================================================================
 Name        : main.c
 Author      : $(author)
 Version     :
 Copyright   : $(copyright)
 Description : main definition
===============================================================================
*/

#if defined (__USE_LPCOPEN)
#if defined(NO_BOARD_LIB)
#include "chip.h"
#else
#include "board.h"
#endif
#endif
#include "FreeRTOS.h"
#include "task.h"
#include <string.h>
#include <stdlib.h>

#include <cr_section_macros.h>

#include <diva_wing5_keyboard.h>

static int readPinStapping();
static int getNodeId();

static char hexdigit(const int value)
{
	return (value < 10) ? (value + '0') : (value < 16) ? (value - 10 + 'A') : 0;
}

static void hex2str(char *s, const unsigned int value, const int digits)
{
	int i;
	for (i = 0; i < digits; ++i)
	{
		s[i] = hexdigit((value>>((digits-1-i)*4))&15);
	}
	s[digits] = 0;
}

/* Transmit a byte array through the UART peripheral (non-blocking) */
int mywrite(LPC_USART_T *pUART, const void *data, int numBytes)
{
	int remains = numBytes;
	uint8_t *p = (uint8_t *) data;

	while (remains > 0)
	{
    	while ((LPC_USART->LSR & UART_LSR_THRE)==0);
   		LPC_USART->THR = (uint32_t)(*p);

		p++;
		--remains;
	}
	return 0;
}

/* LED0 toggle thread */
static void vLEDTask0(void *pvParameters)
{
	while (1)
	{
		//mywrite(LPC_USART, "Hello\r\n", 7);
    	//while ((LPC_USART->LSR & UART_LSR_THRE)==0);
   		//LPC_USART->THR = (uint32_t)'A';
	}
}

static int median5(const int * _values)
{
	int i;
	int unsorted = 1;
	int values[5] = {_values[0], _values[1], _values[2], _values[3], _values[4] };
	while(unsorted)
	{
		unsorted = 0;
		for (i=0; i < 4; ++i)
		{
			if (values[i] > values[i+1])
			{
				const int t = values[i];
				values[i] = values[i+1];
				values[i+1] = t;
				unsorted = 1;
			}
		}
	}
	return values[2];
}

enum { FaderCount = 5 };
static int faderValues[FaderCount][5];
static int faderInPtr[FaderCount];
static int filteredFaderValue[FaderCount];
static uint32_t oldKeyboardState = 0;
static void faderTaskFunction(void *pvParameters)
{
	memset (faderValues, 0, sizeof(faderValues));
	memset (faderInPtr, 0, sizeof(faderInPtr));

	while (1)
	{
		char str[10];
		for (int i = 0; i < 5; ++i)
		{
			faderValues[i][faderInPtr[i]] = Fader_ReadValue(i);
			faderInPtr[i] = (faderInPtr[i] + 1) % 5;
			const int v = median5(faderValues[i]);
			if (abs(v - filteredFaderValue[i]) > 1)
			{
				filteredFaderValue[i] = v;

				str[0] = 'F';
		    	hex2str(&str[1], i, 2);
		    	hex2str(&str[3], filteredFaderValue[i], 3);
		    	str[6] = '\r';
		    	str[7] = '\n';
		    	str[8] = 0;
		    	Board_UARTPutSTR(str);
			}
		}

		if (Keyboard_GetState() != oldKeyboardState)
		{
			const uint32_t keyChange = Keyboard_GetState() ^ oldKeyboardState;
			for (int i = 0; i < 15; ++i)
			{
				if (keyChange & (1<<i))
				{
					str[0] = ((Keyboard_GetState()>>i) & 1) ? 'K' : 'k';
			    	hex2str(&str[1], i, 3);
			    	str[4] = '\r';
			    	str[5] = '\n';
			    	str[6] = 0;
			    	Board_UARTPutSTR(str);
				}
			}
			oldKeyboardState = Keyboard_GetState();
		}

		vTaskDelay(configTICK_RATE_HZ / 50);
	}
}


#if 0
static void faderTestTaskFunction(void *pvParameters)
{
	static int faderValues[FaderCount][5];
	static int faderInPtr[FaderCount];
	static int filteredFaderValue[FaderCount];

	memset (faderValues, 0, sizeof(faderValues));
	memset (faderInPtr, 0, sizeof(faderInPtr));

	int counter = 0;

	while (1)
	{
		char str[30];

		Board_UARTPutSTR("fader:");
		for (int i = 0; i < 5; ++i)
		{
			faderValues[i][faderInPtr[i]] = Fader_ReadValue(i);
			faderInPtr[i] = (faderInPtr[i] + 1) % 5;
			const int v = median5(faderValues[i]);
			if (abs(v - filteredFaderValue[i]) > 1)
			{
				filteredFaderValue[i] = v;
			}

			if (Keyboard_GetState() & (1<<(i*3)))
			{
				Keyboard_SetLed(i, 0, 0xf, filteredFaderValue[i]);
			}
			Keyboard_SetLed(i, 1, 0xf, filteredFaderValue[i]);
			Keyboard_SetLed(i, 2, 0xf, filteredFaderValue[i]);
			if (Keyboard_GetState() & (2<<(i*3)))
			{
				Keyboard_SetLed(i, 3, 0xf, filteredFaderValue[i]);
			}
			if (Keyboard_GetState() & (4<<(i*3)))
			{
				Keyboard_SetLed(i, 4, 0xf, filteredFaderValue[i]);
			}

	    	hex2str(str, filteredFaderValue[i], 3);
	    	str[3] = ' ';
	    	str[4] = 0;
	    	Board_UARTPutSTR(str);

		    vTaskDelay(configTICK_RATE_HZ / 500);
		}
		hex2str(str, Keyboard_GetState(), 8);
		Board_UARTPutSTR(str);

		vTaskDelay(configTICK_RATE_HZ / 100);
		Board_UARTPutSTR(" ");

		vTaskDelay(configTICK_RATE_HZ / 50);
		++counter;
		Board_UARTPutSTR("\r");
	}
}
#endif

static inline int hexchar2int(const char c)
{
	switch (c)
	{
	case 'A'...'F': return c - 'A' + 10;
	case 'a'...'f': return c - 'a' + 10;
	case '0'...'9': return c - '0';
	default:
		break;
	}
	return 0;
}

static int hexstr2int(const char *str, const int count)
{
	int value = 0;
	for (int i = 0; str[i] && i < count; ++i)
	{
		value = (value * 16) + hexchar2int(str[i]);
	}
	return value;
}

/* LED2 toggle thread */
static void commandInterfaceThread(void *pvParameters)
{
	char buffer[20];
	unsigned char count = 0;
	while (1)
	{
		int c = Board_UARTGetChar();
		if (c != EOF)
		{
			if ((c==13) || (c==10))
			{
				if (count > 0)
				{
					buffer[count++] = 0;
					count = 0;
					if (buffer[0] == 'L')
					{
						int blinkMask = hexstr2int(&buffer[1], 1);
						int column = hexstr2int(&buffer[2], 1);
						int row = hexstr2int(&buffer[3], 1);
						int value = hexstr2int(&buffer[4], 3);
						Keyboard_SetLed(column, row, blinkMask, value);
					}
					Board_UARTPutSTR(buffer);
					Board_UARTPutChar('\r');
					Board_UARTPutChar('\n');
				}
			}
			else
			{
				if (count < sizeof(buffer)-1)
				{
					buffer[count++] = c;
				}
			}
		}
		else
		{
			vTaskDelay(configTICK_RATE_HZ/10);
		}
	}
}

/*
 * Configure pinstrapping pins as inputs with pull-ups enabled,
 * read their values and return them after restoring their original configuration.
 * The Pinstrapping pins used are PIO0_2, PIO0_3, PIO0_8 and PIO0_9.
 * These are used as outputs in this design.
 */
static int readPinStapping()
{
#define PINSTRAP_PIN_COUNT (4)
	const uint8_t             pinstrap_ports[PINSTRAP_PIN_COUNT] = {0x02, 0x03, 0x08, 0x09};
	const enum CHIP_IOCON_PIO pio_registers[PINSTRAP_PIN_COUNT]  = {IOCON_PIO0_2,IOCON_PIO0_3,IOCON_PIO0_8,IOCON_PIO0_9};

	uint32_t pinstrapValue = 0;
	uint32_t pio_mode[PINSTRAP_PIN_COUNT];
	bool pio_dir[PINSTRAP_PIN_COUNT];

	for (int i = 0; i < PINSTRAP_PIN_COUNT; ++i)
	{
		const int port = (pinstrap_ports[i]>>4) & 15;
		const int pin = pinstrap_ports[i] & 15;

		// save pin-config state
		pio_mode[i] = LPC_IOCON->REG[pio_registers[i]];
		pio_dir[i] = Chip_GPIO_GetPinDIR(LPC_GPIO, port, pin);

		// switch to input
		Chip_IOCON_PinMuxSet(LPC_IOCON, pio_registers[i], FUNC0|IOCON_DIGMODE_EN|IOCON_MODE_PULLUP);
		Chip_GPIO_SetPinDIRInput(LPC_GPIO, port, pin);

		// read value
		pinstrapValue |= Chip_GPIO_GetPinState(LPC_GPIO, port, pin) ? (1<<i) : 0;

		// restore saved state.
		Chip_IOCON_PinMuxSet(LPC_IOCON, pio_registers[i], pio_mode[i]);
		Chip_GPIO_SetPinDIR(LPC_GPIO, port, pin, pio_dir[i]);
	}
	return pinstrapValue;
}

static int getNodeId()
{
	static int nodeId = -1;
	if (nodeId == -1)
	{
		nodeId = readPinStapping();
	}
	return nodeId;
}

int main(void)
{

#if defined (__USE_LPCOPEN)
#if !defined(NO_BOARD_LIB)
    // Read clock settings and update SystemCoreClock variable
    SystemCoreClockUpdate();
    // Set up and initialize all required blocks and
    // functions related to the board hardware
    Board_Init();
    // Set the LED to the state of "On"
    Board_LED_Set(0, true);
#endif
#endif

    Board_Debug_Init();
    Board_UARTPutSTR("Hello\r\n");

    /*
     * Read node id via pinstrapping.
     * This call reads the pins and caches the result.
     */
    (void)getNodeId();

    /*
     * The Timer is used to control the capture of the Keyboard
     * as well as dimming the LED's.
     */
    Keyboard_Initialize();
    Board_UARTPutSTR("Keyboard OK\r\n");
    Fader_Initialize();
    Board_UARTPutSTR("Fader OK\r\n");


	xTaskCreate(commandInterfaceThread, (signed char *) "vTaskLed2",
				configMINIMAL_STACK_SIZE+32, NULL, (tskIDLE_PRIORITY + 3UL),
				(xTaskHandle *) NULL);

	xTaskCreate(faderTaskFunction, (signed char *) "fader1",
				configMINIMAL_STACK_SIZE+32, NULL, (tskIDLE_PRIORITY + 1UL),
				(xTaskHandle *) NULL);

    Board_UARTPutSTR("FaderTask OK\r\n");

	/* LED0 toggle thread */
#if 0
    xTaskCreate(vLEDTask0, (signed char *) "vTaskLed0",
				configMINIMAL_STACK_SIZE, NULL, (tskIDLE_PRIORITY + 1UL),
				(xTaskHandle *) NULL);
    Board_UARTPutSTR("LedTask OK\r\n");
#endif
	/* Start the scheduler */
	vTaskStartScheduler();

	return 0 ;
}
